// Copyright 2025 RnD Center "ELVEES", JSC

#include <cmath>
#include <malloc.h>
#include <string.h>
#include <stdio.h>

#include <elcore50-matrix-lib/convolution_with_dma_fl16.h>
#include <elcore50-matrix-lib/convolution_small_asm_ker.h>

#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))

#define FL_16

#define EPS 0.001
#define COEF 2

extern "C" void float16_to_float32(uint16_t* src, float* dst, int size);
extern "C" void float32_to_float16(float* src, uint16_t* dst, int size);

void set_data_tensor(Tensor_fl32* tensor);
void tensor_data_copy(Tensor_fl32* src, Tensor_fl32* dst);
void weight_data_copy(Weight_fl32* src, Weight_fl32* dst);


void set_data_weight_dyxc(Weight_fl32* weight);
void weight_dyxc_to_yxcd(const Weight_fl32* input, Weight_fl32* output);
void print_tensor(Tensor_fl32* tensor);

int comparator_fl32(float* src0, float* src1, int size, float eps);

int array_src_b[] =    {  1,   1,   1,   1,   1,   1,   1,   1,   1,   1};
int array_src_h[] =    {112,  17,  17,  28,  28,   8,   8,   8,   8,   8};
int array_src_w[] =    {112,  17,  17,  28,  28,   8,   8,   8,   8,   8};
int array_src_ch[] =   { 28, 128, 128, 128, 128, 384, 384, 384, 384, 384};
int array_kerX[] =     {  3,   1,   7,   3,   3,   3,   3,   3,   3,   3};
int array_kerY[] =     {  3,   7,   1,   3,   3,   1,   1,   1,   1,   1};
int array_dst_b[] =    {  1,   1,   1,   1,   1,   1,   1,   1,   1,   1};
int array_dst_h[] =    {112,  17,  17,  28,  28,   8,   8,   8,   8,   8};
int array_dst_w[] =    {112,  17,  17,  28,  28,   8,   8,   8,   8,   8};
int array_dst_ch[] =   { 28, 128, 128, 128, 256, 384,  32,  64,  96, 160};
int array_stride_X[] = {  1,   1,   1,   1,   1,   1,   1,   1,   1,   1};
int array_stride_Y[] = {  1,   1,   1,   1,   1,   1,   1,   1,   1,   1};
int bias_flags[] =     {  1,   1,   1,   0,   1,   1,   1,   1,   1,   1};
int array_padX[] =     {  1,   0,   3,   1,   1,   1,   1,   1,   1,   1};
int array_padY[] =     {  1,   3,   0,   1,   1,   0,   0,   0,   0,   0};
int array_padW[] =     {  1,   0,   3,   1,   1,   1,   1,   1,   1,   1};
int array_padH[] =     {  1,   3,   0,   1,   1,   0,   0,   0,   0,   0};

int main()
{
    uint32_t l2dc_val = *L2DC;
    flush_all_caches();
    *L2DC &= 0xFFFFFFBF;

    printf("| func_name        | input (b, h, w, ch)    | bias | store mode |"
           " opt mul/tic | status |\n");
    printf("|                  | ker (X, Y), str (X, Y) |      |            |"
           "             |        |\n");
    printf("|                  | output (b, h, w, ch)   |      |            |"
           "             |        |\n");
    printf("-----------------------------------------------------------------"
           "-----------------------\n");

    change_L1DC_ctrl();
    Store_version st_ver = STORE_NONE;

    for (int st_v = 0; st_v < 1; ++st_v){
        st_ver = STORE_NONE;


    for (int indx = 0; indx < 10; ++indx)
    {
        printf("| convolution fl16 | (%1d %4d %4d %4d)     |    %d |",
               array_src_b[indx], array_src_h[indx], array_src_w[indx],
               array_src_ch[indx], bias_flags[indx]);

        if (st_ver == STORE_NONE)
            printf("      none  | ");
        else if (st_ver == STORE_RELU)
            printf("      relu  | ");
        else
            printf("      relu6 | ");

        int result = 0;
        int padY, padX, padH, padW;
        //                        dH   dW   sH  sW  sC    dC
        Debug_data tile_params = { 4,   4,   4,  4,  8,  128};

        padY = padX = padH = padW = 0;
        padX = array_padX[indx];
        padY = array_padY[indx];
        padW = array_padW[indx];
        padH = array_padH[indx];

        // шаги ядра
        int strideY = array_stride_Y[indx];
        int strideX = array_stride_X[indx];

        //printf("INDX = %d\n", indx);

        int src_ch_for_ref_ver = array_src_ch[indx];
        if (array_src_ch[indx] % 8)
        {
            src_ch_for_ref_ver += 8 - (array_src_ch[indx] % 8);
        }

                                //data, b, h,  w, c
        Tensor_fl32 input = {NULL, array_src_b[indx], array_src_h[indx], array_src_w[indx], array_src_ch[indx]};
        input.data =
            (float*)memalign(64, sizeof(float) * input.batch * input.height * input.width * input.channel * COEF);
        set_data_tensor(&input);

        // тензор кратного размера
        Tensor_fl32 input_for_ref_ver =
            {NULL, array_src_b[indx], array_src_h[indx], array_src_w[indx], src_ch_for_ref_ver};
        input_for_ref_ver.data = (float*)memalign(64, sizeof(float) * input_for_ref_ver.batch * input_for_ref_ver.height
                                                      * input_for_ref_ver.width * input_for_ref_ver.channel * COEF);

        tensor_data_copy(&input, &input_for_ref_ver);

        Tensor_fl16 input_fl16_for_ref = {NULL, input.batch, input.height, input.width, input_for_ref_ver.channel};
        input_fl16_for_ref.data = (uint16_t*)memalign(64, input_fl16_for_ref.batch * input_fl16_for_ref.height
                                                          * input_fl16_for_ref.width * input_fl16_for_ref.channel
                                                          * sizeof(uint16_t) * COEF);
        float32_to_float16(
            input_for_ref_ver.data, input_fl16_for_ref.data,
            input_fl16_for_ref.batch * input_fl16_for_ref.height * input_fl16_for_ref.width
            * input_fl16_for_ref.channel);


        Tensor_fl16 input_fl16= {NULL, input.batch, input.height, input.width, input.channel};
        input_fl16.data = (uint16_t*)memalign(64, input_fl16.batch * input_fl16.height
                                                  * input_fl16.width * input_fl16.channel * sizeof(uint16_t) * COEF);
        float32_to_float16(input.data, input_fl16.data, input.batch * input.height * input.width * input.channel);

        Tensor_fl32 output_ref = {NULL, array_dst_b[indx], array_dst_h[indx], array_dst_w[indx], array_dst_ch[indx]};
        output_ref.data = (float*)memalign(64, sizeof(float) * output_ref.batch * output_ref.height
                                                        * output_ref.width * output_ref.channel * COEF);
        memset(output_ref.data, 0, sizeof(float) * output_ref.batch * output_ref.height
                                                        * output_ref.width * output_ref.channel * COEF);


        Tensor_fl32 output_opt = {NULL, output_ref.batch, output_ref.height, output_ref.width, output_ref.channel};
        output_opt.data = (float*)memalign(64, sizeof(float) * output_opt.batch * output_opt.height
                                                        * output_opt.width * output_opt.channel * COEF);
        memset(output_opt.data, 0, sizeof(float) * output_opt.batch * output_opt.height
                                                        * output_opt.width * output_opt.channel * COEF);


        //printf("OUTPUT: b=%d  h=%d  w=%d  c=%d\n", output_ref.batch, output_ref.height, output_ref.width, output_ref.channel);

        Tensor_fl16 output_opt_fl16 = {NULL, output_ref.batch, output_ref.height, output_ref.width, output_ref.channel};
        output_opt_fl16.data =
            (uint16_t*)calloc(output_opt_fl16.batch * output_opt_fl16.height * output_opt_fl16.width
                              * output_opt_fl16.channel, sizeof(uint16_t)  * COEF);



                                 //data, X, Y
        Weight_fl32 weight_dyxc = {NULL, array_kerX[indx], array_kerY[indx], input.channel, output_ref.channel};
        Weight_fl32 weight_yxcd = {NULL, weight_dyxc.kernelX, weight_dyxc.kernelY, input.channel, output_ref.channel};
        weight_dyxc.data = (float*)memalign(64, sizeof(float) * weight_dyxc.kernelX * weight_dyxc.kernelY
                                                               * input.channel * output_ref.channel);
        set_data_weight_dyxc(&weight_dyxc);
        weight_yxcd.data = (float*)memalign(64, sizeof(float) * weight_yxcd.kernelX * weight_yxcd.kernelY
                                                               * input.channel * output_ref.channel);
        weight_dyxc_to_yxcd(&weight_dyxc, &weight_yxcd);

        //printf("KERNEL: kerX=%d  kerY=%d  strideX=%d  strideY=%d\n", weight_dyxc.kernelX, weight_dyxc.kernelY, strideX, strideY);

        Weight_fl32 weight_yxcd_ref_ver =
            {NULL, weight_dyxc.kernelX, weight_dyxc.kernelY, input_for_ref_ver.channel, output_ref.channel};
        weight_yxcd_ref_ver.data = (float*)memalign(64, sizeof(float) * weight_yxcd.kernelX * weight_yxcd.kernelY
                                                               * input_for_ref_ver.channel * output_ref.channel);

        Weight_fl16 weight_yxcd_fl16_ref_ver =
            {NULL, weight_dyxc.kernelX, weight_dyxc.kernelY, input_for_ref_ver.channel, output_ref.channel};
        weight_yxcd_fl16_ref_ver.data =
            (uint16_t*)memalign(64, sizeof(uint16_t) * weight_yxcd.kernelX * weight_yxcd.kernelY
                                    * input_for_ref_ver.channel * output_ref.channel);

        weight_data_copy(&weight_yxcd, &weight_yxcd_ref_ver);

        float32_to_float16(
            weight_yxcd_ref_ver.data, weight_yxcd_fl16_ref_ver.data,
            weight_yxcd.kernelX * weight_yxcd.kernelY * input_for_ref_ver.channel * output_ref.channel);


        Weight_fl16 weight_yxcd_fl16 =
        {NULL, weight_dyxc.kernelX, weight_dyxc.kernelY, input.channel, output_ref.channel};
        weight_yxcd_fl16.data = (uint16_t*)memalign(64, sizeof(uint16_t) * weight_yxcd.kernelX * weight_yxcd.kernelY
                                                        * input.channel * output_ref.channel);
        float32_to_float16(weight_yxcd.data, weight_yxcd_fl16.data, weight_yxcd.kernelX * weight_yxcd.kernelY
                                                                    * input.channel * output_ref.channel);

        float *bias = (float*)memalign(64, output_ref.channel * sizeof(float));

        for (int i = 0; i < output_ref.channel; ++i)
        {
            bias[i] = (bias_flags[indx]) ? (rand()%3) : 0;
        }

        // растяжение свертки
        int dilationY = 1;
        int dilationX = 1;
        // границы изображения дописанные

        //printf("padX = %d padY = %d padW = %d padH = %d\n", padX, padY, padW, padH);

        int group = 1;

    #ifdef DEBUG
        //printf("run ref func\n");
    #endif

        //printf("RUN REF VER\n");
        ref_convolution(input.data, input.batch, input.height, input.width, input.channel,
        weight_dyxc.kernelY, weight_dyxc.kernelX, dilationY, dilationX,
        strideY, strideX, padY, padX, padH, padW, group,
        weight_dyxc.data, bias, output_ref.data, output_ref.channel);



        memset(output_opt_fl16.data, 0, output_opt_fl16.batch * output_opt_fl16.height
                                        * output_opt_fl16.width * output_opt_fl16.channel * sizeof(uint16_t) * COEF);

        flush_all_caches();

        memset(output_opt_fl16.data, 0, output_opt_fl16.batch * output_opt_fl16.height
                                        * output_opt_fl16.width * output_opt_fl16.channel * sizeof(uint16_t) * COEF);

        if (st_ver != STORE_NONE)
        {
            for (int i = 0; i < output_ref.height; ++i)
            {
                for (int j = 0; j < output_ref.width; ++j)
                {
                    for (int k = 0; k < output_ref.channel; ++k)
                    {
                        int res_indx = i * output_ref.width * output_ref.channel + j * output_ref.channel + k;
                        if (output_ref.data[res_indx] < 0.0)
                            output_ref.data[res_indx] = 0.0;

                        if (st_ver == STORE_RELU6)
                        {
                            if (output_ref.data[res_indx] > 6.0)
                                output_ref.data[res_indx] = 6.0;
                        }
                    }
                }
            }
        }

        /////////////////////////////////////////
        ConvLocalMemBufFl16 local_buf;
        VDMAChain chain_A, chain_B, chain_st_C, chain_ld_C, chain_init_vec;
        MatMulDmaChains4 chains = {&chain_A, &chain_B, &chain_st_C, &chain_ld_C, &chain_init_vec};

        int initcnt = 32;
        initchain(chains.chain_A, initcnt);
        initchain(chains.chain_B, initcnt);
        initchain(chains.chain_st_C, initcnt);
        initchain(chains.chain_ld_C, initcnt);
        initchain(chains.chain_init_vec, initcnt);

        int tic[2], instr[2], func_tic, func_instr;
        uint16_t** zeros = (uint16_t**)malloc(sizeof(uint16_t*));

        conv_size_selector(input_fl16.batch, input_fl16.height, input_fl16.width, input_fl16.channel,
                           weight_yxcd_fl16.kernelY, weight_yxcd_fl16.kernelX,
                           output_opt_fl16.height, output_opt_fl16.width, output_opt_fl16.channel,
                           &local_buf,
                           strideY, strideX,
                           2, 2, 2, bias_flags[indx], 524288,
                           &(local_buf.offset_A));

        //printf("tile src: (%d[%d] %d %d) tile dst: (%d %d %d)\n", local_buf.buf_srcH, local_buf.offset_A, local_buf.buf_srcW, local_buf.buf_srcC,
        //                                                          local_buf.buf_dstH, local_buf.buf_dstW, local_buf.buf_dstC);

        Conv_version conv_ver;
        if (local_buf.buf_dstC > 64)
        {
            conv_ver = VER_4x4x128;
        }
        else if (local_buf.buf_dstC > 32)
        {
            conv_ver = VER_4x8x64;
        }
        else
        {
            conv_ver = VER_8x8x32;
        }


        if (local_buf.buf_srcC >= input_fl16.channel)
        {
            /////////////////// версия без шага по srcC ///////////////////
            init_chain_dma_convolution_small_ker_2A2B2C(input_fl16.data, input_fl16.batch, input_fl16.height, input_fl16.width, input_fl16.channel,
                                                       weight_yxcd_fl16.kernelY, weight_yxcd_fl16.kernelX, dilationY, dilationX,
                                                       strideY, strideX, output_opt_fl16.height, output_opt_fl16.width, group,
                                                       padX, padW, padY, padH,
                                                       weight_yxcd_fl16.data, bias, output_opt_fl16.data, output_opt_fl16.channel, &local_buf, &chains, zeros, bias_flags[indx],
                                                       (uint16_t*)&__local_mem, 524288);
            count_tics(tic, instr);
            convolution_small_ker_2A2B2C(input_fl16.data, input_fl16.batch, input_fl16.height, input_fl16.width, input_fl16.channel,
                                        weight_yxcd_fl16.kernelY, weight_yxcd_fl16.kernelX, dilationY, dilationX,
                                        strideY, strideX, output_opt_fl16.height, output_opt_fl16.width, group,
                                        weight_yxcd_fl16.data, bias, output_opt_fl16.data, output_opt_fl16.channel,
                                        &local_buf, &chains, NULL, st_ver, bias_flags[indx], conv_ver);

            count_tics(&tic[1], &instr[1]);
            //////////////////////////////////////////////////////////////
        }
        else
        {
            /////////////////// версия с шагом по srcC ///////////////////
            float** inter_res = (float**)malloc(sizeof(float*));

            //printf("RUN INIT\n");
            init_chain_dma_convolution_small_ker_2A2B2C_srcC_ver(input_fl16.data, input_fl16.batch, input_fl16.height, input_fl16.width, input_fl16.channel,
                                                       weight_yxcd_fl16.kernelY, weight_yxcd_fl16.kernelX, dilationY, dilationX,
                                                       strideY, strideX, output_opt_fl16.height, output_opt_fl16.width, group,
                                                       padX, padW, padY, padH,
                                                       weight_yxcd_fl16.data, bias, output_opt_fl16.data, output_opt_fl16.channel, &local_buf, &chains, zeros, bias_flags[indx],
                                                       (uint16_t*)&__local_mem, 524288, inter_res, tile_params);
            //printf("RUN OPT\n-----------------------------\n");
            count_tics(tic, instr);
            convolution_small_ker_2A2B2C_srcC_ver(input_fl16.data, input_fl16.batch, input_fl16.height, input_fl16.width, input_fl16.channel,
                                        weight_yxcd_fl16.kernelY, weight_yxcd_fl16.kernelX, dilationY, dilationX,
                                        strideY, strideX, output_opt_fl16.height, output_opt_fl16.width, group,
                                        weight_yxcd_fl16.data, bias, output_opt_fl16.data, output_opt_fl16.channel,
                                        &local_buf, &chains, NULL, st_ver, bias_flags[indx], *inter_res, conv_ver);

            count_tics(&tic[1], &instr[1]);
            free(*inter_res);
            free(inter_res);
        }
        ///////////////////////////////////////////////////////////////

        destroychain(chains.chain_A);
        destroychain(chains.chain_B);
        destroychain(chains.chain_st_C);
        destroychain(chains.chain_ld_C);
        destroychain(chains.chain_init_vec);

        flush_all_caches();

        func_tic = tic[1] - tic[0];
        func_instr = instr[1] - instr[0];
        //printf("tic: %d %f mul/tic (%d) instr: %f mul/instr %d\n", func_tic, 1.0 * array_dst_h[indx] * array_dst_w[indx] * array_src_ch[indx] * array_kerX[indx] * array_kerY[indx] * array_dst_ch[indx] / func_tic, 0, 1.0 * array_dst_h[indx] * array_dst_w[indx] * array_src_ch[indx] * array_dst_ch[indx] / func_instr, func_instr);

        printf("   %7.3f  | ", (float)array_dst_h[indx] * array_dst_w[indx] * array_src_ch[indx] * array_kerX[indx] * array_kerY[indx] * array_dst_ch[indx] / func_tic);

        //printf("zero free\n");
        free(*zeros);
        free(zeros);
        //printf("inter_res free\n");
        //printf("end free\n");

        /////////////////////////////////////////


        float16_to_float32(output_opt_fl16.data, output_opt.data, output_opt_fl16.batch * output_opt_fl16.height
                                                 * output_opt_fl16.width * output_opt_fl16.channel * COEF);


        //print_tensor(&output_ref);
        //print_tensor(&output_opt);

        /*for (int i = 0; i < 512; ++i)
        {
            printf("%f ", output_opt.data[i]);
        }*/

        result += comparator_fl32(output_ref.data, output_opt.data, output_ref.batch * output_ref.height
                                                                   * output_ref.width * output_ref.channel * COEF, (float)EPS);

        //printf("result = %d\n", result);

        if (!result)
            printf("passed |\n");
        else
            printf("failed |\n");

        printf("|                  | (%1d %2d) (%1d %2d)          |      |            |             |        |\n", array_kerX[indx], array_kerY[indx], array_stride_X[indx], array_stride_Y[indx]);
        printf("|                  | (%1d %4d %4d %4d)     |      |            |             |        |\n", array_dst_b[indx], array_dst_h[indx], array_dst_w[indx], array_dst_ch[indx]);
        printf("----------------------------------------------------------------------------------------\n");

        //if (result)
        //    printf("ERROR\n");
        //for (int i = 0; i < 9; ++i)
        //    printf("%f\n", output_ref.data[i]);
        //assert(result == 0);



        free(weight_yxcd_fl16.data);
        free(output_opt_fl16.data);
        free(input_fl16.data);

        free(output_opt.data);
        free(input.data);
        free(output_ref.data);
        free(weight_dyxc.data);
        free(weight_yxcd.data);
        free(bias);
    }
}

    *L2DC = l2dc_val;
    return 0;
}

void set_data_tensor(Tensor_fl32* tensor)
{
    int indx = 0;
    for (int b = 0; b < tensor->batch; ++b)
    {
        for (int h = 0; h < tensor->height; ++h)
        {
            for (int w = 0; w < tensor->width; ++w)
            {
                for (int ch = 0; ch < tensor->channel; ++ch)
                {
                    indx = b * tensor->height * tensor->width * tensor->channel
                           + h * tensor->width * tensor->channel
                           + w * tensor->channel + ch;

                    tensor->data[indx] = rand()%3 - 1;

                }
            }
        }
    }
}

void tensor_data_copy(Tensor_fl32* src, Tensor_fl32* dst)
{
    memset(dst->data, 0, dst->batch * dst->height * dst->width * dst->channel * sizeof(float));

    int indx_src, indx_dst;
    for (int b = 0; b < src->batch; ++b)
    {
        for (int h = 0; h < src->height; ++h)
        {
            for (int w = 0; w < src->width; ++w)
            {
                for (int ch = 0; ch < src->channel; ++ch)
                {
                    indx_src = b * src->height * src->width * src->channel
                           + h * src->width * src->channel
                           + w * src->channel + ch;

                    indx_dst = b * dst->height * dst->width * dst->channel
                           + h * dst->width * dst->channel
                           + w * dst->channel + ch;

                    dst->data[indx_dst] = src->data[indx_src];
                }
            }
        }
    }
}

void set_data_weight_dyxc(Weight_fl32* weight)
{
    int indx = 0;
    for (int b = 0; b < weight->dstC; ++b)
    {
        for (int h = 0; h < weight->kernelY; ++h)
        {
            for (int w = 0; w < weight->kernelX; ++w)
            {
                for (int ch = 0; ch < weight->srcC; ++ch)
                {
                    indx = b * weight->kernelY * weight->kernelX * weight->srcC
                           + h * weight->kernelX * weight->srcC
                           + w * weight->srcC + ch;
                    weight->data[indx] = rand()%3 - 1;
                }
            }
        }
    }
}


void weight_dyxc_to_yxcd(const Weight_fl32* input, Weight_fl32* output)
{
    int indx_in = 0;
    int indx_out = 0;
    for (int h = 0; h < output->kernelY; ++h)
    {
        for (int w = 0; w < output->kernelX; ++w)
        {
            for (int ch = 0; ch < output->srcC; ++ch)
            {
                for (int b = 0; b < output->dstC; ++b)
                {
                    indx_out = h * output->dstC * output->kernelX * output->srcC
                               + w * output->srcC * output->dstC
                               + ch * output->dstC + b;
                    indx_in = b * output->kernelY * output->kernelX * output->srcC
                              + h * output->kernelX * output->srcC
                              + w * output->srcC + ch;
                    output->data[indx_out] = input->data[indx_in];
                }
            }
        }
    }
}

void weight_data_copy(Weight_fl32* src, Weight_fl32* dst)
{
    memset(dst->data, 0, dst->dstC * dst->srcC * dst->kernelX * dst->kernelY * sizeof(float));

    int indx_src = 0;
    int indx_dst = 0;
    for (int h = 0; h < src->kernelY; ++h)
    {
        for (int w = 0; w < src->kernelX; ++w)
        {
            for (int ch = 0; ch < src->srcC; ++ch)
            {
                for (int ch_d = 0; ch_d < src->dstC; ++ch_d)
                {
                    indx_src = ch_d + ch * src->dstC + w * src->dstC * src->srcC
                                    + h * src->dstC * src->srcC * src->kernelX;

                    indx_dst = ch_d + ch * dst->dstC + w * dst->dstC * dst->srcC
                                    + h * dst->dstC * dst->srcC * dst->kernelX;


                    dst->data[indx_dst] = src->data[indx_src];
                }
            }
        }
    }
}

void print_tensor(Tensor_fl32* tensor)
{
    for (int b = 0; b < tensor->batch * COEF; ++b)
    {
        for (int h = 0; h <  tensor->height; ++h)
        {
            for (int ch = 0; ch <  1; ++ch)
            {
                for (int w = 0; w <  tensor->width; ++w)
                {
                    printf("%5.1f ", tensor->data[b * tensor->height * tensor->channel * tensor->width
                                                  + h * tensor->channel * tensor->width
                                                  + w * tensor->channel + ch]);
                }
                printf("     ");
            }
            printf("\n");
        }
        printf("\n");
    }
}

int comparator_fl32(float* src0, float* src1, int size, float eps)
{
    int err_counter = 0;
    for (int i = 0; i < size; ++i)
    {
        if (src0[i] != src0[i] || src1[i] != src1[i])
        {
            printf("NAN: [%d] %f != %f\n", i, src0[i], src1[i]);
            return 11;
        }


        if (src0[i] != 0 && src1[i] != 0 && err_counter < 10)
        {
            if ((fabs(src0[i] - src1[i]) / MIN(fabs(src0[i]), fabs(src1[i]))) > eps)
            {
                printf("Error: [%d] %f != %f err = %f\n", i, src0[i], src1[i], (fabs(src0[i] - src1[i]) / MIN(fabs(src0[i]), fabs(src1[i]))));
                ++err_counter;
            }
        }
        else
        {
            if (MAX(fabs(src0[i]), fabs(src1[i])) > eps && err_counter < 10)
            {
                printf("Error: [%d] %f != %f\n", i, src0[i], src1[i]);
                ++err_counter;
            }
        }
    }

    return err_counter;
}
